@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2006 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -24,6 +24,7 @@
 #include <linux/mm.h>
 #include <linux/xattr.h>
 #include <linux/posix_acl.h>
+#include <linux/security.h>
 
 #include "gfs.h"
 #include "acl.h"
@@ -42,6 +43,48 @@
 #include "trans.h"
 #include "unlinked.h"
 
+/**
+ * gfs_security_init -
+ * @dip:
+ * @ip:
+ *
+ * Returns: errno
+ */
+
+static int
+gfs_security_init(struct gfs_inode *dip, struct gfs_inode *ip)
+{
+	int err;
+	size_t len;
+	void *value;
+	char *name;
+	struct gfs_ea_request er;
+
+	err = security_inode_init_security(ip->i_vnode, dip->i_vnode,
+					   &name, &value, &len);
+
+	if (err) {
+		if (err == -EOPNOTSUPP)
+			return 0;
+		return err;
+	}
+
+	memset(&er, 0, sizeof(struct gfs_ea_request));
+
+	er.er_type = GFS_EATYPE_SECURITY;
+	er.er_name = name;
+	er.er_data = value;
+	er.er_name_len = strlen(name);
+	er.er_data_len = len;
+
+	err = gfs_ea_set_i(ip, &er);
+
+	kfree(value);
+	kfree(name);
+
+	return err;
+}
+
 /**
  * gfs_create - Create a file
  * @dir: The directory in which to create the file
@@ -75,7 +118,7 @@ gfs_create(struct inode *dir, struct dentry *dentry,
 		if (!error)
 			break;
 		else if (error != -EEXIST ||
-			 (nd->intent.open.flags & O_EXCL)) {
+			 (nd && (nd->intent.open.flags & O_EXCL))) {
 			gfs_holder_uninit(&d_gh);
 			return error;
 		}
@@ -104,15 +147,20 @@ gfs_create(struct inode *dir, struct dentry *dentry,
 		gfs_alloc_put(dip);
 	}
 
-	gfs_glock_dq_uninit(&d_gh);
-	gfs_glock_dq_uninit(&i_gh);
-
 	inode = gfs_iget(ip, CREATE);
 	gfs_inode_put(ip);
 
 	if (!inode)
 		return -ENOMEM;
 
+	error = gfs_security_init(dip, ip);
+
+	gfs_glock_dq_uninit(&d_gh);
+	gfs_glock_dq_uninit(&i_gh);
+
+	if (error)
+		return error;
+
 	d_instantiate(dentry, inode);
 	if (new)
 		mark_inode_dirty(inode);
@@ -142,17 +190,17 @@ lookup_cdpn_sub_at(struct gfs_sbd *sdp, struct dentry *dentry)
 	parent = dget_parent(dentry);
 
 	if (gfs_filecmp(&dentry->d_name, "@hostname", 9))
-		new = lookup_one_len(system_utsname.nodename,
+		new = lookup_one_len(init_utsname()->nodename,
 				     parent,
-				     strlen(system_utsname.nodename));
+				     strlen(init_utsname()->nodename));
 	else if (gfs_filecmp(&dentry->d_name, "@mach", 5))
-		new = lookup_one_len(system_utsname.machine,
+		new = lookup_one_len(init_utsname()->machine,
 				     parent,
-				     strlen(system_utsname.machine));
+				     strlen(init_utsname()->machine));
 	else if (gfs_filecmp(&dentry->d_name, "@os", 3))
-		new = lookup_one_len(system_utsname.sysname,
+		new = lookup_one_len(init_utsname()->sysname,
 				     parent,
-				     strlen(system_utsname.sysname));
+				     strlen(init_utsname()->sysname));
 	else if (gfs_filecmp(&dentry->d_name, "@uid", 4))
 		new = lookup_one_len(buf,
 				     parent,
@@ -165,8 +213,8 @@ lookup_cdpn_sub_at(struct gfs_sbd *sdp, struct dentry *dentry)
 		new = lookup_one_len(buf,
 				     parent,
 				     sprintf(buf, "%s_%s",
-					     system_utsname.machine,
-					     system_utsname.sysname));
+					     init_utsname()->machine,
+					     init_utsname()->sysname));
 	else if (gfs_filecmp(&dentry->d_name, "@jid", 4))
 		new = lookup_one_len(buf,
 				     parent,
@@ -201,17 +249,17 @@ lookup_cdpn_sub_brace(struct gfs_sbd *sdp, struct dentry *dentry)
 	parent = dget_parent(dentry);
 
 	if (gfs_filecmp(&dentry->d_name, "{hostname}", 10))
-		new = lookup_one_len(system_utsname.nodename,
+		new = lookup_one_len(init_utsname()->nodename,
 				     parent,
-				     strlen(system_utsname.nodename));
+				     strlen(init_utsname()->nodename));
 	else if (gfs_filecmp(&dentry->d_name, "{mach}", 6))
-		new = lookup_one_len(system_utsname.machine,
+		new = lookup_one_len(init_utsname()->machine,
 				     parent,
-				     strlen(system_utsname.machine));
+				     strlen(init_utsname()->machine));
 	else if (gfs_filecmp(&dentry->d_name, "{os}", 4))
-		new = lookup_one_len(system_utsname.sysname,
+		new = lookup_one_len(init_utsname()->sysname,
 				     parent,
-				     strlen(system_utsname.sysname));
+				     strlen(init_utsname()->sysname));
 	else if (gfs_filecmp(&dentry->d_name, "{uid}", 5))
 		new = lookup_one_len(buf,
 				     parent,
@@ -224,8 +272,8 @@ lookup_cdpn_sub_brace(struct gfs_sbd *sdp, struct dentry *dentry)
 		new = lookup_one_len(buf,
 				     parent,
 				     sprintf(buf, "%s_%s",
-					     system_utsname.machine,
-					     system_utsname.sysname));
+					     init_utsname()->machine,
+					     init_utsname()->sysname));
 	else if (gfs_filecmp(&dentry->d_name, "{jid}", 5))
 		new = lookup_one_len(buf,
 				     parent,
@@ -595,11 +643,16 @@ gfs_symlink(struct inode *dir, struct dentry *dentry, const char *symname)
 	gfs_unlinked_unlock(sdp, dip->i_alloc->al_ul);
 	gfs_alloc_put(dip);
 
+	inode = gfs_iget(ip, CREATE);
+	gfs_inode_put(ip);
+
+	error = gfs_security_init(dip, ip);
+
 	gfs_glock_dq_uninit(&d_gh);
 	gfs_glock_dq_uninit(&i_gh);
 
-	inode = gfs_iget(ip, CREATE);
-	gfs_inode_put(ip);
+	if (error)
+		return error;
 
 	if (!inode)
 		return -ENOMEM;
@@ -689,15 +742,20 @@ gfs_mkdir(struct inode *dir, struct dentry *dentry, int mode)
 	gfs_unlinked_unlock(sdp, dip->i_alloc->al_ul);
 	gfs_alloc_put(dip);
 
-	gfs_glock_dq_uninit(&d_gh);
-	gfs_glock_dq_uninit(&i_gh);
-
 	inode = gfs_iget(ip, CREATE);
 	gfs_inode_put(ip);
 
 	if (!inode)
 		return -ENOMEM;
 
+	error = gfs_security_init(dip, ip);
+
+	gfs_glock_dq_uninit(&d_gh);
+	gfs_glock_dq_uninit(&i_gh);
+
+	if (error)
+		return error;
+
 	d_instantiate(dentry, inode);
 	mark_inode_dirty(inode);
 
@@ -861,11 +919,16 @@ gfs_mknod(struct inode *dir, struct dentry *dentry, int mode, dev_t dev)
 	gfs_unlinked_unlock(sdp, dip->i_alloc->al_ul);
 	gfs_alloc_put(dip);
 
+	inode = gfs_iget(ip, CREATE);
+	gfs_inode_put(ip);
+
+	error = gfs_security_init(dip, ip);
+
 	gfs_glock_dq_uninit(&d_gh);
 	gfs_glock_dq_uninit(&i_gh);
 
-	inode = gfs_iget(ip, CREATE);
-	gfs_inode_put(ip);
+	if (error)
+		return error;
 
 	if (!inode)
 		return -ENOMEM;
@@ -1187,7 +1250,7 @@ gfs_readlink(struct dentry *dentry, char *user_buf, int user_size)
  * Returns: 0 on success or error code
  */
 
-static int
+static void *
 gfs_follow_link(struct dentry *dentry, struct nameidata *nd)
 {
 	struct gfs_inode *ip = get_v2ip(dentry->d_inode);
@@ -1204,7 +1267,7 @@ gfs_follow_link(struct dentry *dentry, struct nameidata *nd)
 			kfree(buf);
 	}
 
-	return error;
+	return ERR_PTR(error);
 }
 
 /**
@@ -1221,65 +1284,7 @@ gfs_follow_link(struct dentry *dentry, struct nameidata *nd)
 static int
 gfs_permission_i(struct inode *inode, int mask, struct nameidata *nd)
 {
-	int mode = inode->i_mode;
-      
-	/* Nobody gets write access to a read-only fs */
-	if ((mask & MAY_WRITE) &&
-	    IS_RDONLY(inode) &&
-	    (S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode)))
-		return -EROFS;
-
-	/* Nobody gets write access to an immutable file */
-	if ((mask & MAY_WRITE) && IS_IMMUTABLE(inode))
-		return -EACCES;
-
-	if (current->fsuid == inode->i_uid)
-		mode >>= 6;
-	else if (IS_POSIXACL(inode)) {
-		struct posix_acl *acl = NULL;
-		int error;
-
-		/* The access ACL cannot grant access if the group class
-		   permission bits don't contain all requested permissions. */
-		if (((mode >> 3) & mask & S_IRWXO) != mask)
-			goto check_groups;
-
-		error = gfs_acl_get(get_v2ip(inode), TRUE, &acl);
-		if (error)
-			return error;
-
-		if (acl) {
-			int error = posix_acl_permission(inode, acl, mask);
-			posix_acl_release(acl);
-			if (error == -EACCES)
-				goto check_capabilities;
-			return error;
-		} else
-			goto check_groups;
-	} else {
-	check_groups:
-		if (in_group_p(inode->i_gid))
-			mode >>= 3;
-	}
-
-	if ((mode & mask & S_IRWXO) == mask)
-		return 0;
-      
- check_capabilities:
-	/* Allowed to override Discretionary Access Control? */
-	if (!(mask & MAY_EXEC) ||
-	    (inode->i_mode & S_IXUGO) ||
-	    S_ISDIR(inode->i_mode))
-		if (capable(CAP_DAC_OVERRIDE))
-			return 0;
-
-	/* Read and search granted if capable(CAP_DAC_READ_SEARCH) */
-	if (capable(CAP_DAC_READ_SEARCH) &&
-	    ((mask == MAY_READ) ||
-	     (S_ISDIR(inode->i_mode) && !(mask & MAY_WRITE))))
-		return 0;
-
-	return -EACCES;
+	return generic_permission(inode, mask, gfs_check_acl);
 }
 
 /**
@@ -1335,7 +1340,35 @@ gfs_setattr(struct dentry *dentry, struct iattr *attr)
 
 	atomic_inc(&sdp->sd_ops_inode);
 
+	/* Bugzilla 203170: we'll have the same deadlock as described 
+	 * in bugzilla 173912 if
+	 * 1. without RHEL4's DIO_CLUSTER_LOCKING, and
+	 * 2. we come down to this line of code from do_truncate()
+	 *    where i_sem(i_mutex) and i_alloc_sem have been taken, and
+	 * 3. grab the exclusive glock here.
+	 * To avoid this to happen, i_alloc_sem must be dropped and trust
+	 * be put into glock that it can carry the same protection. 
+	 *
+	 * One issue with dropping i_alloc_sem is that the gfs_setattr() 
+	 * can be invoked from other code path without this sempaphore. 
+	 * We'll need a new rwsem function that can "up" the semaphore 
+	 * only when it is needed. Before that happens (will research the 
+	 * possibility), i_alloc_sem (now) is a meaningless lock within 
+	 * GFS. If it is ever been used by other non-directIO code, this
+	 * hack will fall apart.
+	 *
+	 * wcheng@redhat.com 10/14/06  
+	 */ 
+	if (attr->ia_valid & ATTR_SIZE) {
+		up_write(&dentry->d_inode->i_alloc_sem); 
+	}
+	
 	error = gfs_glock_nq_init(ip->i_gl, LM_ST_EXCLUSIVE, 0, &i_gh);
+
+	if (attr->ia_valid & ATTR_SIZE) {
+		down_write(&dentry->d_inode->i_alloc_sem); 
+	}
+	
 	if (error)
 		return error;
 
@@ -1480,8 +1513,7 @@ gfs_getattr(struct vfsmount *mnt, struct dentry *dentry, struct kstat *stat)
 	atomic_inc(&ip->i_sbd->sd_ops_inode);
 
 	error = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY, &gh);
-	if (!error)
-	{
+	if (!error) {
 		generic_fillattr(inode, stat);
 		gfs_glock_dq_uninit(&gh);
 	}
