@@ -11,12 +11,12 @@
 *******************************************************************************
 ******************************************************************************/
 
+#include <linux/kthread.h>
 #include <linux/sched.h>
 #include <linux/slab.h>
 #include <linux/smp_lock.h>
 #include <linux/spinlock.h>
 #include <asm/semaphore.h>
-#include <linux/completion.h>
 #include <linux/buffer_head.h>
 #include <linux/vmalloc.h>
 #include <linux/statfs.h>
@@ -112,51 +112,25 @@ gfs_put_super(struct super_block *sb)
 	up(&sdp->sd_freeze_lock);
 
 	/*  Kill off the inode thread  */
-	down(&sdp->sd_thread_lock);
-	clear_bit(SDF_INODED_RUN, &sdp->sd_flags);
-	wake_up_process(sdp->sd_inoded_process);
-	up(&sdp->sd_thread_lock);
-	wait_for_completion(&sdp->sd_thread_completion);
+	kthread_stop(sdp->sd_inoded_process);
 
 	/*  Kill off the quota thread  */
-	down(&sdp->sd_thread_lock);
-	clear_bit(SDF_QUOTAD_RUN, &sdp->sd_flags);
-	wake_up_process(sdp->sd_quotad_process);
-	up(&sdp->sd_thread_lock);
-	wait_for_completion(&sdp->sd_thread_completion);
+	kthread_stop(sdp->sd_quotad_process);
 
 	/*  Kill off the log thread  */
-	down(&sdp->sd_thread_lock);
-	clear_bit(SDF_LOGD_RUN, &sdp->sd_flags);
-	wake_up_process(sdp->sd_logd_process);
-	up(&sdp->sd_thread_lock);
-	wait_for_completion(&sdp->sd_thread_completion);
+	kthread_stop(sdp->sd_logd_process);
 
 	/*  Kill off the recoverd thread  */
-	down(&sdp->sd_thread_lock);
-	clear_bit(SDF_RECOVERD_RUN, &sdp->sd_flags);
-	wake_up_process(sdp->sd_recoverd_process);
-	up(&sdp->sd_thread_lock);
-	wait_for_completion(&sdp->sd_thread_completion);
+	kthread_stop(sdp->sd_recoverd_process);
 
 	/*  Kill off the glockd threads  */
-	clear_bit(SDF_GLOCKD_RUN, &sdp->sd_flags);
-	wake_up(&sdp->sd_reclaim_wchan);
 	while (sdp->sd_glockd_num--)
-		wait_for_completion(&sdp->sd_thread_completion);
+		kthread_stop(sdp->sd_glockd_process[sdp->sd_glockd_num]);
 
 	/*  Kill off the scand thread  */
-	down(&sdp->sd_thread_lock);
-	clear_bit(SDF_SCAND_RUN, &sdp->sd_flags);
-	wake_up_process(sdp->sd_scand_process);
-	up(&sdp->sd_thread_lock);
-	wait_for_completion(&sdp->sd_thread_completion);
+	kthread_stop(sdp->sd_scand_process);
 
 	if (!test_bit(SDF_ROFS, &sdp->sd_flags)) {
-		gfs_log_flush(sdp);
-		gfs_quota_sync(sdp);
-		gfs_quota_sync(sdp);
-
 		error = gfs_make_fs_ro(sdp);
 		if (error)
 			gfs_io_error(sdp);
@@ -231,6 +205,9 @@ gfs_write_super_lockfs(struct super_block *sb)
 	struct gfs_sbd *sdp = get_v2sdp(sb);
 	int error;
 
+	if (test_bit(SDF_SHUTDOWN, &sdp->sd_flags))
+		return;
+
 	atomic_inc(&sdp->sd_ops_super);
 
 	for (;;) {
@@ -322,6 +299,7 @@ static int
 gfs_remount_fs(struct super_block *sb, int *flags, char *data)
 {
 	struct gfs_sbd *sdp = get_v2sdp(sb);
+	struct gfs_tune *gt = &sdp->sd_tune;
 	int error = 0;
 	struct gfs_args *args;
 
@@ -331,7 +309,7 @@ gfs_remount_fs(struct super_block *sb, int *flags, char *data)
 	if (!args)
 		return -ENOMEM;
 
-	error = gfs_make_args(data, args, FALSE);
+	error = gfs_make_args(data, args, TRUE);
 	if (error) {
 		printk("GFS: can't parse remount arguments\n");
 		goto out;
@@ -362,6 +340,25 @@ gfs_remount_fs(struct super_block *sb, int *flags, char *data)
 		}
 	}
 
+	if (args->ar_noquota) {
+		if (sdp->sd_args.ar_noquota == FALSE)
+			printk("GFS: remounting without quota\n");
+		sdp->sd_args.ar_noquota = TRUE;
+		spin_lock(&gt->gt_spin);
+		gt->gt_quota_enforce = 0;
+		gt->gt_quota_account = 0;
+		spin_unlock(&gt->gt_spin);
+	}
+	else {
+		if (sdp->sd_args.ar_noquota == TRUE)
+			printk("GFS: remounting with quota\n");
+		sdp->sd_args.ar_noquota = FALSE;
+		spin_lock(&gt->gt_spin);
+		gt->gt_quota_enforce = 1;
+		gt->gt_quota_account = 1;
+		spin_unlock(&gt->gt_spin);
+	}
+
 	/*  Don't let the VFS update atimes.  GFS handles this itself. */
 	*flags |= MS_NOATIME | MS_NODIRATIME;
 
@@ -444,6 +441,8 @@ gfs_show_options(struct seq_file *s, struct vfsmount *mnt)
 		seq_printf(s, ",num_glockd=%u", args->ar_num_glockd);
 	if (args->ar_posix_acls)
 		seq_printf(s, ",acl");
+	if (args->ar_noquota)
+		seq_printf(s, ",noquota");
 	if (args->ar_suiddir)
 		seq_printf(s, ",suiddir");
 
