@@ -74,17 +74,37 @@ int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 				return 1;
 			}
 		} else {
-			error = pass->check_dentry(ip, dent, prev, bh, filename, update,
-									   count, pass->private);
-			if(error < 0) {
-				stack;
-				return -1;
+			if (!de.de_inum.no_addr && first) { /* reverse sentinel */
+				log_debug("First dirent is a Sentinel (place holder).\n");
+				/* Swap the two to silently make it a proper sentinel */
+				de.de_inum.no_addr = de.de_inum.no_formal_ino;
+				de.de_inum.no_formal_ino = 0;
+				gfs2_dirent_out(&de, (char *)dent);
+				*update = 1; /* Mark dirent buffer as modified */
+				first = 0;
+			}
+			else {
+				error = pass->check_dentry(ip, dent, prev, bh, filename,
+							   update, count, pass->private);
+				if(error < 0) {
+					stack;
+					return -1;
+				}
+				/*if(error > 0) {
+				  return 1;
+				  }*/
 			}
-			/*if(error > 0) {
-			  return 1;
-			  }*/
 		}
 
+		if (de.de_rec_len < sizeof(struct gfs2_dirent)) {
+			log_err("Entry %" PRIu64 "(0x%"
+				PRIx64 ") of directory %" PRIu64 "(0x%"
+				PRIx64 ") is corrupt, skipping.\n",
+				bh->b_blocknr, bh->b_blocknr,
+				ip->i_di.di_num.no_addr,
+				ip->i_di.di_num.no_addr);
+                        break;
+                }
 		if ((char *)dent + de.de_rec_len >= bh_end){
 			log_debug("Last entry processed.\n");
 			break;
@@ -151,21 +171,48 @@ int check_leaf(struct gfs2_inode *ip, int *update, struct metawalk_fxns *pass)
 				}
 			}
 
+			*update = not_updated;
 			lbh = bread(sbp, leaf_no);
 			gfs2_leaf_in(&leaf, lbh->b_data);
 
+			/*
+			 * Early versions of GFS2 had an endianess bug in the kernel
+			 * that set lf_dirent_format to cpu_to_be16(GFS2_FORMAT_DE).
+			 * This was fixed to use cpu_to_be32(), but we should check
+			 * for incorrect values and replace them with the correct value. */
+
+			if (leaf.lf_dirent_format == (GFS2_FORMAT_DE << 16)) {
+				log_debug("incorrect lf_dirent_format at leaf #%" PRIu64 "\n", leaf_no);
+				leaf.lf_dirent_format = GFS2_FORMAT_DE;
+				gfs2_leaf_out(&leaf, lbh->b_data);
+				log_debug("Fixing lf_dirent_format.\n");
+				*update = updated;
+			}
+
+			/* Make sure it's really a leaf. */
+			if (leaf.lf_header.mh_type != GFS2_METATYPE_LF) {
+				log_err("Inode %" PRIu64 " (0x%"
+					PRIx64 ") points to bad leaf "
+					PRIu64 " (0x%" PRIx64 ").\n",
+					ip->i_di.di_num.no_addr, leaf_no);
+				brelse(lbh, *update);
+				break;
+			}
 			exp_count = (1 << (ip->i_di.di_depth - leaf.lf_depth));
 			log_debug("expected count %u - di_depth %u, leaf depth %u\n",
 					  exp_count, ip->i_di.di_depth, leaf.lf_depth);
+
 			if(pass->check_dentry &&
 			   S_ISDIR(ip->i_di.di_mode)) {
 				error = check_entries(ip, lbh, index, DIR_EXHASH, update,
 									  &count, pass);
 
 				/* Since the buffer possibly got
-				   updated directly, release it now,
-				   and grab it again later if we need it */
-				brelse(lbh, not_updated);
+				 * updated directly, release it now,
+				 * and grab it again later if we need it. */
+
+				brelse(lbh, *update);
+
 				if(error < 0) {
 					stack;
 					return -1;
@@ -175,9 +222,8 @@ int check_leaf(struct gfs2_inode *ip, int *update, struct metawalk_fxns *pass)
 					return 1;
 
 				if(update && (count != leaf.lf_entries)) {
-					enum update_flags f;
+					enum update_flags f = not_updated;
 
-					f = not_updated;
 					lbh = bread(sbp, leaf_no);
 					gfs2_leaf_in(&leaf, lbh->b_data);
 
@@ -199,7 +245,7 @@ int check_leaf(struct gfs2_inode *ip, int *update, struct metawalk_fxns *pass)
 				 * compare it against leaf->lf_entries */
 				break;
 			} else {
-				brelse(lbh, not_updated);
+				brelse(lbh, *update);
 				if(!leaf.lf_next)
 					break;
 				leaf_no = leaf.lf_next;
@@ -219,6 +265,7 @@ static int check_eattr_entries(struct gfs2_inode *ip,
 	uint64_t *ea_data_ptr = NULL;
 	int i;
 	int error = 0;
+	uint32_t offset = (uint32_t)sizeof(struct gfs2_meta_header);
 
 	if(!pass->check_eattr_entry) {
 		return 0;
@@ -259,11 +306,11 @@ static int check_eattr_entries(struct gfs2_inode *ip,
 				}
 			}
 		}
-		if(ea_hdr->ea_flags & GFS2_EAFLAG_LAST){
-			/* FIXME: better equal the end of the block */
+		offset += be32_to_cpu(ea_hdr->ea_rec_len);
+		if(ea_hdr->ea_flags & GFS2_EAFLAG_LAST ||
+		   offset >= ip->i_sbd->sd_sb.sb_bsize || ea_hdr->ea_rec_len == 0){
 			break;
 		}
-		/* FIXME: be sure this doesn't go beyond the end */
 		ea_hdr_prev = ea_hdr;
 		ea_hdr = (struct gfs2_ea_header *)
 			((char *)(ea_hdr) +
