@@ -801,6 +801,13 @@ int pass1(struct gfs2_sbd *sbp)
 			if (gfs2_next_rg_meta(rgd, &block, first))
 				break;
 			warm_fuzzy_stuff(block);
+			if (fsck_abort) /* if asked to abort */
+				return 0;
+			if (skip_this_pass) {
+				printf("Skipping pass 1 is not a good idea.\n");
+				skip_this_pass = FALSE;
+				fflush(stdout);
+			}
 			bh = bread(sbp, block);
 
 			if (scan_meta(sbp, bh, block)) {
