@@ -18,6 +18,7 @@
 #include <sys/ioctl.h>
 #include <sys/stat.h>
 #include <sys/utsname.h>
+#include <sys/time.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <net/if.h>
@@ -46,6 +47,16 @@
 static int control_fd = -1;
 extern int our_nodeid;
 static int plocks_online = 0;
+extern int message_flow_control_on;
+extern int no_plock;
+
+extern uint32_t plock_rate_limit;
+uint32_t plock_read_count;
+uint32_t plock_recv_count;
+uint32_t plock_rate_delays;
+struct timeval plock_read_time;
+struct timeval plock_recv_time;
+struct timeval plock_rate_last;
 
 static SaCkptHandleT ckpt_handle;
 static SaCkptCallbacksT callbacks = { 0, 0 };
@@ -274,12 +285,23 @@ int setup_plocks(void)
 	SaAisErrorT err;
 	int rv;
 
+	plock_read_count = 0;
+	plock_recv_count = 0;
+	plock_rate_delays = 0;
+	gettimeofday(&plock_read_time, NULL);
+	gettimeofday(&plock_recv_time, NULL);
+	gettimeofday(&plock_rate_last, NULL);
+
+	if (no_plock)
+		goto control;
+
 	err = saCkptInitialize(&ckpt_handle, &callbacks, &version);
 	if (err == SA_AIS_OK)
 		plocks_online = 1;
 	else
 		log_error("ckpt init error %d - plocks unavailable", err);
 
+ control:
 	rv = open_control();
 	if (rv)
 		return rv;
@@ -289,17 +311,56 @@ int setup_plocks(void)
 	return control_fd;
 }
 
+static unsigned long time_diff_ms(struct timeval *begin, struct timeval *end)
+{
+	struct timeval result;
+	timersub(end, begin, &result);
+	return (result.tv_sec * 1000) + (result.tv_usec / 1000);
+}
+
 int process_plocks(void)
 {
 	struct mountgroup *mg;
 	struct gdlm_plock_info info;
 	struct gdlm_header *hd;
+	struct timeval now;
 	char *buf;
 	int len, rv;
 
+	/* Don't send more messages while the cpg message queue is backed up */
+
+	if (message_flow_control_on) {
+		update_flow_control_status();
+		if (message_flow_control_on)
+			return -EBUSY;
+	}
+
+	/* Every N ops we check how long it's taken to do those N ops.
+	   If it's less than 1000 ms, we don't take any more. */
+
+	if (plock_rate_limit && plock_read_count &&
+	    !(plock_read_count % plock_rate_limit)) {
+		gettimeofday(&now, NULL);
+		if (time_diff_ms(&plock_rate_last, &now) < 1000) {
+			plock_rate_delays++;
+			return -EBUSY;
+		}
+		plock_rate_last = now;
+	}
+
 	memset(&info, 0, sizeof(info));
 
-	rv = read(control_fd, &info, sizeof(info));
+	rv = do_read(control_fd, &info, sizeof(info));
+	if (rv < 0) {
+		log_debug("process_plocks: read error %d fd %d\n",
+			  errno, control_fd);
+		return 0;
+	}
+
+	if (!plocks_online) {
+		rv = -ENOSYS;
+		goto fail;
+	}
 
 	mg = find_mg_id(info.fsid);
 	if (!mg) {
@@ -316,6 +377,18 @@ int process_plocks(void)
 		  info.nodeid, info.pid, info.owner,
 		  info.wait);
 
+	/* report plock rate and any delays since the last report */
+	plock_read_count++;
+	if (!(plock_read_count % 1000)) {
+		gettimeofday(&now, NULL);
+		log_group(mg, "plock_read_count %u time %us delays %u",
+			  plock_read_count,
+			  (unsigned) (now.tv_sec - plock_read_time.tv_sec),
+			  plock_rate_delays);
+		plock_read_time = now;
+		plock_rate_delays = 0;
+	}
+
 	len = sizeof(struct gdlm_header) + sizeof(struct gdlm_plock_info);
 	buf = malloc(len);
 	if (!buf) {
@@ -850,7 +923,7 @@ static void do_get(struct mountgroup *mg, struct gdlm_plock_info *in)
 }
 
 /* When mg members receive our options message (for our mount), one of them
-   saves all plock state received to that point in a checkpoint and then sounds
+   saves all plock state received to that point in a checkpoint and then sends
    us our journals message.  We know to retrieve the plock state from the
    checkpoint when we receive our journals message.  Any plocks messages that
    arrive between seeing our options message and our journals message needs to
@@ -863,6 +936,7 @@ void _receive_plock(struct mountgroup *mg, char *buf, int len, int from)
 {
 	struct gdlm_plock_info info;
 	struct gdlm_header *hd = (struct gdlm_header *) buf;
+	struct timeval now;
 	int rv = 0;
 
 	memcpy(&info, buf + sizeof(struct gdlm_header), sizeof(info));
@@ -877,6 +951,14 @@ void _receive_plock(struct mountgroup *mg, char *buf, int len, int from)
 		  info.nodeid, info.pid, info.owner,
 		  info.wait);
 
+	plock_recv_count++;
+	if (!(plock_recv_count % 1000)) {
+		gettimeofday(&now, NULL);
+		log_group(mg, "plock_recv_count %u time %us", plock_recv_count,
+			  (unsigned) (now.tv_sec - plock_recv_time.tv_sec));
+		plock_recv_time = now;
+	}
+
 	if (info.optype == GDLM_PLOCK_OP_GET && from != our_nodeid)
 		return;
 
@@ -928,8 +1010,6 @@ void process_saved_plocks(struct mountgroup *mg)
 {
 	struct save_msg *sm, *sm2;
 
-	mg->save_plocks = 0;
-
 	if (list_empty(&mg->saved_messages))
 		return;
 
@@ -1376,6 +1456,7 @@ void retrieve_plocks(struct mountgroup *mg)
 	if (mg->low_nodeid == our_nodeid) {
 		/* we're the new low nodeid, will be master */
 		log_group(mg, "retrieve_plocks: unlink ckpt from old master");
+		mg->cp_handle = (uint64_t) h;
 		_unlink_checkpoint(mg, &name);
 	} else
 		saCkptCheckpointClose(h);
