@@ -213,7 +213,7 @@ send_ret(msgctx_t *ctx, char *name, int ret, int orig_request)
 
 	
 void
-send_response(int ret, request_t *req)
+send_response(int ret, int nodeid, request_t *req)
 {
 	SmMessageSt msg, *msgp = &msg;
 
@@ -226,7 +226,10 @@ send_response(int ret, request_t *req)
 	msgp->sm_data.d_action = req->rr_orig_request;
 	strncpy(msgp->sm_data.d_svcName, req->rr_group,
 		sizeof(msgp->sm_data.d_svcName));
-	msgp->sm_data.d_svcOwner = my_id(); /* XXX Broken */
+	if (!nodeid)
+		msgp->sm_data.d_svcOwner = my_id();
+	else 
+		msgp->sm_data.d_svcOwner = nodeid;
 	msgp->sm_data.d_ret = ret;
 
 	swab_SmMessageSt(msgp);
@@ -245,7 +248,7 @@ set_rg_state(char *name, rg_state_t *svcblk)
 	char res[256];
 #ifndef OPENAIS
 	cluster_member_list_t *membership;
-	int ret;
+	int ret, tries = 0;
 #endif
 
 	if (name)
@@ -257,11 +260,16 @@ set_rg_state(char *name, rg_state_t *svcblk)
 		return -1;
 	return 0;
 #else
-	membership = member_list();
-	ret = vf_write(membership, VFF_IGN_CONN_ERRORS, res, svcblk,
-       		       sizeof(*svcblk));
-	free_member_list(membership);
-	return ret;
+	do {
+		/* Retry up to 3 times just in case members transition
+		   while we're trying to commit something */
+		membership = member_list();
+		ret = vf_write(membership, VFF_IGN_CONN_ERRORS, res, svcblk,
+       		       	       sizeof(*svcblk));
+		free_member_list(membership);
+	} while (ret == VFR_TIMEOUT && ++tries < 3);
+
+	return (ret==VFR_OK?0:-1);
 #endif
 }
 
@@ -452,10 +460,6 @@ svc_advise_stop(rg_state_t *svcStatus, char *svcName, int req)
 		break;
 		
 	case RG_STATE_STOPPING:
-		printf("Service %s is stopping right now\n", svcName);
-		ret = 2;
-		break;
-
 	case RG_STATE_STARTED:
 	case RG_STATE_CHECK:
 	case RG_STATE_STARTING:
@@ -561,6 +565,7 @@ svc_advise_stop(rg_state_t *svcStatus, char *svcName, int req)
  *			1 = START service - return whatever it returns.
  *			2 = DO NOT start service, return 0
  *			3 = DO NOT start service, return RG_EAGAIN
+ *			4 = DO NOT start service, return RG_ERUN
  */
 int
 svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
@@ -585,7 +590,7 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 			clulog(LOG_DEBUG,
 			       "RG %s is already running locally\n", svcName);
 			 */
-			ret = 2;
+			ret = 4;
 			break;
 		}
 
@@ -597,7 +602,7 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 			       svcName,
 			       memb_id_to_name(membership,svcStatus->rs_owner));
 			 */
-			ret = 2;
+			ret = 4;
 			break;
 		}
 
@@ -655,6 +660,7 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 		break;
 
 	case RG_STATE_STOPPED:
+	case RG_STATE_ERROR:
 		/* Don't actually enable if the RG is locked! */
 		if (rg_locked()) {
 			ret = 3;
@@ -686,7 +692,6 @@ svc_advise_start(rg_state_t *svcStatus, char *svcName, int req)
 		       svcName);
 		break;
 
-	case RG_STATE_ERROR:
 	default:
 		clulog(LOG_ERR,
 		       "#44: Cannot start RG %s: Invalid State %d\n",
@@ -738,6 +743,9 @@ svc_start(char *svcName, int req)
 	case 3:
 		rg_unlock(&lockp);
 		return RG_EAGAIN;
+	case 4:
+		rg_unlock(&lockp);
+		return RG_ERUN;
 	default:
 		break;
 	}
@@ -858,6 +866,10 @@ svc_migrate(char *svcName, int target)
 	rg_unlock(&lockp);
        
 	ret = group_migrate(svcName, target);
+
+	if (ret)
+		svc_fail(svcName);
+
 	return ret;
 }
 
@@ -1193,7 +1205,7 @@ relocate_service(char *svcName, int request, uint32_t target)
 	swab_SmMessageSt(&msg_relo);
 
 	/* Send relocate message to the other node */
-	if (msg_send(&ctx, &msg_relo, sizeof (SmMessageSt)) !=
+	if (msg_send(&ctx, &msg_relo, sizeof (SmMessageSt)) < 
 	    sizeof (SmMessageSt)) {
 		clulog(LOG_ERR,
 		       "#59: Error sending relocate request to member #%d\n",
@@ -1280,7 +1292,7 @@ handle_relocate_req(char *svcName, int request, int preferred_target,
 		    int *new_owner)
 {
 	cluster_member_list_t *allowed_nodes, *backup = NULL;
-	uint32_t target = preferred_target, me = my_id();
+	int target = preferred_target, me = my_id();
 	int ret, x;
 	
 	/*
@@ -1296,7 +1308,7 @@ handle_relocate_req(char *svcName, int request, int preferred_target,
 			return RG_EFORWARD;
 	}
 
-	if (preferred_target != 0) {
+	if (preferred_target >= 0) {
 
 		allowed_nodes = member_list();
 		/*
@@ -1368,7 +1380,7 @@ handle_relocate_req(char *svcName, int request, int preferred_target,
 		//count_resource_groups(allowed_nodes);
 	}
 
-	if (preferred_target != 0)
+	if (preferred_target >= 0)
 		memb_mark_down(allowed_nodes, preferred_target);
 	memb_mark_down(allowed_nodes, me);
 
@@ -1483,8 +1495,8 @@ handle_start_req(char *svcName, int req, int *new_owner)
 	/* 
 	   If services are locked, return the error 
 	  */
-	if (ret == RG_EAGAIN)
-		return RG_EAGAIN;
+	if (ret == RG_EAGAIN || ret == RG_ERUN)
+		return ret;
 
 	/*
 	 * If we succeeded, then we're done.
